@@ -216,6 +216,17 @@ bool FMStatic::copy(const QList<QUrl> &urls, const QUrl &destinationDir)
 #endif
 }
 
+bool FMStatic::group(const QList<QUrl>& urls, const QUrl& destinationDir, const QString &name)
+{
+    if(!FMStatic::fileExists(destinationDir.toString()+"/"+name))
+    {
+        QDir(destinationDir.toLocalFile()).mkdir(name);
+    }
+    
+   return FMStatic::cut(urls, destinationDir, name);
+}
+
+
 bool FMStatic::cut(const QList<QUrl> &urls, const QUrl &where)
 {
     return FMStatic::cut(urls, where, QString());
@@ -227,7 +238,7 @@ bool FMStatic::cut(const QList<QUrl> &urls, const QUrl &where, const QString &na
     QUrl _where = where;
     if (!name.isEmpty())
         _where = QUrl(where.toString() + "/" + name);
-
+    
     auto job = KIO::move(urls, _where, KIO::HideProgressInfo);
     job->start();
 
@@ -368,7 +379,7 @@ void FMStatic::openUrl(const QUrl &url)
 {
 #ifdef KIO_AVAILABLE
     KIO::OpenUrlJob *job = new KIO::OpenUrlJob(url);
-    job->setShowOpenOrExecuteDialog(true);
+    job->setRunExecutables(true);
     job->start();
 #else
 
@@ -595,5 +606,3 @@ const FMH::MODEL FMStatic::getFileInfo(const KFileItem &kfile)
             {
                     return FMStatic::PATHTYPE_SCHEME_NAME[url.scheme()];
         }
-
-
