@@ -17,30 +17,70 @@
  *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
  */
 
-#include <QCoreApplication>
 #include <QMimeDatabase>
-// #include <KFileMetaData/UserMetaData>
+
+#include <QFileInfo>
+#include <QDateTime>
+
+#include <QSqlError>
+#include <QSqlQuery>
+#include <QSqlRecord>
+
+#include <MauiKit/Core/fmh.h>
+
 #include "tagging.h"
 #include "fmstatic.h"
+#include "tagdb.h"
 
 Tagging *Tagging::m_instance = nullptr;
 
-Tagging::Tagging()
-    : TAGDB()
+Tagging::~Tagging()
 {
-    this->setApp();
 }
 
-const QVariantList Tagging::get(const QString &queryTxt, std::function<bool(QVariantMap &item)> modifier)
+Tagging::Tagging() : QObject()
+{
+    this->setApp();
+    connect(qApp, &QCoreApplication::aboutToQuit, [this]()
+    {
+        qDebug() << "Lets remove Tagging singleton instance";
+        
+        qDeleteAll(m_dbs);
+        m_dbs.clear();
+        
+        delete m_instance;
+        m_instance = nullptr;
+    });
+}
+
+TAGDB * Tagging::db()
+{
+    if(m_dbs.contains(QThread::currentThreadId()))
+    {
+        qDebug() << "Using existing TAGGINGDB instance";
+        
+        return m_dbs[QThread::currentThreadId()];
+    }
+    
+    qDebug() << "Creating new TAGGINGDB instance";
+    
+    auto new_db = new TAGDB;
+    m_dbs.insert(QThread::currentThreadId(), new_db);
+    return new_db;
+}
+
+const QVariantList Tagging::get(const QString &queryTxt, std::function<bool(QVariantMap &item)> modifier) 
 {
     QVariantList mapList;
 
-    auto query = this->getQuery(queryTxt);
+    auto query = this->db()->getQuery(queryTxt);
 
     if (query.exec()) {
+        const auto keys = FMH::MODEL_NAME.keys();
+
         while (query.next()) {
             QVariantMap data;
-            for (const auto &key : FMH::MODEL_NAME.keys()) {
+            for (const auto &key : keys) {
                 
                 if (query.record().indexOf(FMH::MODEL_NAME[key]) > -1) {
                     data[FMH::MODEL_NAME[key]] = query.value(FMH::MODEL_NAME[key]).toString();
@@ -49,8 +89,11 @@ const QVariantList Tagging::get(const QString &queryTxt, std::function<bool(QVar
 
             if (modifier) {
                 if (!modifier(data))
+                {
                     continue;
+                }
             }
+
             mapList << data;
         }
 
@@ -62,16 +105,16 @@ const QVariantList Tagging::get(const QString &queryTxt, std::function<bool(QVar
     return mapList;
 }
 
-bool Tagging::tagExists(const QString &tag, const bool &strict)
+bool Tagging::tagExists(const QString &tag, const bool &strict) 
 {
-    return !strict ? this->checkExistance(TAG::TABLEMAP[TAG::TABLE::TAGS], FMH::MODEL_NAME[FMH::MODEL_KEY::TAG], tag)
-                   : this->checkExistance(QString("select t.tag from APP_TAGS where t.org = '%1' and t.tag = '%2'")
-                                              .arg(this->appOrg, tag));
+    return !strict ? this->db()->checkExistance(TAG::TABLEMAP[TAG::TABLE::TAGS], FMH::MODEL_NAME[FMH::MODEL_KEY::TAG], tag)
+    : this->db()->checkExistance(QString("select t.tag from APP_TAGS where t.org = '%1' and t.tag = '%2'")
+            .arg(this->appOrg, tag));
 }
 
-bool Tagging::urlTagExists(const QString &url, const QString &tag)
+bool Tagging::urlTagExists(const QString &url, const QString &tag) 
 {
-  return this->checkExistance(QString("select * from TAGS_URLS where url = '%1' and tag = '%2'").arg(url, tag));
+    return this->db()->checkExistance(QString("select * from TAGS_URLS where url = '%1' and tag = '%2'").arg(url, tag));
 }
 
 void Tagging::setApp()
@@ -99,20 +142,20 @@ bool Tagging::tag(const QString &tag, const QString &color, const QString &comme
         {FMH::MODEL_NAME[FMH::MODEL_KEY::COMMENT], comment},
     };
     
-    this->insert(TAG::TABLEMAP[TAG::TABLE::TAGS], tag_map);
+    this->db()->insert(TAG::TABLEMAP[TAG::TABLE::TAGS], tag_map);
     
     const QVariantMap appTag_map {
-        {FMH::MODEL_NAME[FMH::MODEL_KEY::TAG], tag}, 
-        {FMH::MODEL_NAME[FMH::MODEL_KEY::ORG], this->appOrg}, 
+        {FMH::MODEL_NAME[FMH::MODEL_KEY::TAG], tag},
+        {FMH::MODEL_NAME[FMH::MODEL_KEY::ORG], this->appOrg},
         {FMH::MODEL_NAME[FMH::MODEL_KEY::ADDDATE], QDateTime::currentDateTime().toString(Qt::TextDate)}};
-        
-        if (this->insert(TAG::TABLEMAP[TAG::TABLE::APP_TAGS], appTag_map)) {
-            setTagIconName(tag_map);
-            emit this->tagged(tag_map);
-            return true;
-        }
-        
-        return false;
+
+        if (this->db()->insert(TAG::TABLEMAP[TAG::TABLE::APP_TAGS], appTag_map)) {
+        setTagIconName(tag_map);
+        emit this->tagged(tag_map);
+        return true;
+    }
+
+    return false;
 }
 
 bool Tagging::tagUrl(const QString &url, const QString &tag, const QString &color, const QString &comment)
@@ -130,13 +173,13 @@ bool Tagging::tagUrl(const QString &url, const QString &tag, const QString &colo
                              {FMH::MODEL_NAME[FMH::MODEL_KEY::ADDDATE], QDateTime::currentDateTime()},
                              {FMH::MODEL_NAME[FMH::MODEL_KEY::COMMENT], comment}};
 
-    if(this->insert(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], tag_url_map))
+                             if(this->db()->insert(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], tag_url_map))
     {
         qDebug() << "tagging url" << url <<tag;
         emit this->urlTagged(url, myTag);
         
-//         auto fileMetaData = KFileMetaData::UserMetaData(QUrl::fromUserInput(url).toLocalFile());        
-//         fileMetaData.setTags({tag});
+        //         auto fileMetaData = KFileMetaData::UserMetaData(QUrl::fromUserInput(url).toLocalFile());
+        //         fileMetaData.setTags({tag});
         
         return true;
     }
@@ -146,9 +189,9 @@ bool Tagging::tagUrl(const QString &url, const QString &tag, const QString &colo
 
 bool Tagging::updateUrlTags(const QString &url, const QStringList &tags, const bool &strict)
 {
-    this->removeUrlTags(url, strict);    
-   
-    for (const auto &tag : qAsConst(tags))
+    this->removeUrlTags(url, strict);
+
+    for (const auto &tag : std::as_const(tags))
     {
         this->tagUrl(url, tag);
     }
@@ -156,15 +199,15 @@ bool Tagging::updateUrlTags(const QString &url, const QStringList &tags, const b
     return true;
 }
 
-bool Tagging::updateUrl(const QString &url, const QString &newUrl)
+bool Tagging::updateUrl(const QString &url, const QString &newUrl) 
 {
-    return this->update(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], {{FMH::MODEL_KEY::URL, newUrl}}, {{FMH::MODEL_NAME[FMH::MODEL_KEY::URL], url}});
+    return this->db()->update(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], {{FMH::MODEL_KEY::URL, newUrl}}, {{FMH::MODEL_NAME[FMH::MODEL_KEY::URL], url}});
 }
 
-QVariantList Tagging::getUrlsTags(const bool &strict) //all used tags, meaning, all tags that are used with an url in tags_url table
+QVariantList Tagging::getUrlsTags(const bool &strict)  //all used tags, meaning, all tags that are used with an url in tags_url table
 {
     const auto query = !strict ? QString("select distinct t.* from TAGS t inner join TAGS_URLS turl where t.tag = turl.tag") :
-    QString("select distinct t.* from TAGS t inner join APP_TAGS at on at.tag = t.tag inner join TAGS_URLS turl on t.tag = turl.tag where at.org = '%1'").arg(this->appOrg);    
+                                 QString("select distinct t.* from TAGS t inner join APP_TAGS at on at.tag = t.tag inner join TAGS_URLS turl on t.tag = turl.tag where at.org = '%1'").arg(this->appOrg);
 
     return this->get(query, &setTagIconName);
 }
@@ -175,7 +218,7 @@ bool Tagging::setTagIconName(QVariantMap &item)
     return true;
 }
 
-QVariantList Tagging::getAllTags(const bool &strict)
+QVariantList Tagging::getAllTags(const bool &strict) 
 {
     return !strict ? this->get("select * from tags", &setTagIconName)
                    : this->get(QString("select t.* from TAGS t inner join APP_TAGS at on t.tag = at.tag where at.org = '%1'").arg(this->appOrg),
@@ -190,7 +233,7 @@ QVariantList Tagging::getUrls(const QString &tag, const bool &strict, const int
                                        "inner join TAGS_URLS turl on turl.tag = t.tag "
                                        "where at.org = '%1' and turl.mime like '%4%' "
                                        "and t.tag = '%2' limit %3")
-                                   .arg(this->appOrg, tag, QString::number(limit), mimeType),
+                               .arg(this->appOrg, tag, QString::number(limit), mimeType),
                                modifier);
 }
 
@@ -199,7 +242,7 @@ QVariantList Tagging::getUrlTags(const QString &url, const bool &strict)
     return !strict ? this->get(QString("select distinct turl.*, t.color, t.comment as tagComment from tags t inner join TAGS_URLS turl on turl.tag = t.tag where turl.url  = '%1'").arg(url))
                    : this->get(QString("select distinct t.* from TAGS t inner join APP_TAGS at on t.tag = at.tag inner join TAGS_URLS turl on turl.tag = t.tag "
                                        "where at.org = '%1' and turl.url = '%2'")
-                                   .arg(this->appOrg, url));
+                               .arg(this->appOrg, url));
 }
 
 bool Tagging::removeUrlTags(const QString &url, const bool &strict) // same as removing the url from the tags_urls
@@ -212,12 +255,23 @@ bool Tagging::removeUrlTag(const QString &url, const QString &tag)
 {
     qDebug() << "Remove url tag" << url << tag;
     FMH::MODEL data {{FMH::MODEL_KEY::URL, url}, {FMH::MODEL_KEY::TAG, tag}};
-    return this->remove(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], data);
+    if(this->db()->remove(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], data))
+    {
+        emit this->urlTagRemoved(tag, url);
+        return true;
+    }
+    
+    return false;
 }
 
 bool Tagging::removeUrl(const QString &url)
 {
-    return this->remove(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], {{FMH::MODEL_KEY::URL, url}});
+    if(this->db()->remove(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], {{FMH::MODEL_KEY::URL, url}}))
+    {
+        emit this->urlRemoved(url);
+    }
+    
+    return false;
 }
 
 bool Tagging::app()
@@ -230,26 +284,26 @@ bool Tagging::app()
         {FMH::MODEL_NAME[FMH::MODEL_KEY::COMMENT], this->appComment},
     };
 
-   return this->insert(TAG::TABLEMAP[TAG::TABLE::APPS], app_map);
+    return this->db()->insert(TAG::TABLEMAP[TAG::TABLE::APPS], app_map);
 }
 
 bool Tagging::removeTag(const QString& tag)
 {
-     FMH::MODEL data1 {{FMH::MODEL_KEY::TAG, tag}};
+    FMH::MODEL data1 {{FMH::MODEL_KEY::TAG, tag}};
     
-     if(this->remove(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], data1))
+    if(this->db()->remove(TAG::TABLEMAP[TAG::TABLE::TAGS_URLS], data1))
     {
-             FMH::MODEL data2 {{FMH::MODEL_KEY::TAG, tag}, {FMH::MODEL_KEY::ORG, this->appOrg}};
+        FMH::MODEL data2 {{FMH::MODEL_KEY::TAG, tag}, {FMH::MODEL_KEY::ORG, this->appOrg}};
 
-        if(this->remove(TAG::TABLEMAP[TAG::TABLE::APP_TAGS], data2))
+        if(this->db()->remove(TAG::TABLEMAP[TAG::TABLE::APP_TAGS], data2))
         {
-            if(this->remove(TAG::TABLEMAP[TAG::TABLE::TAGS], data1))
+            if(this->db()->remove(TAG::TABLEMAP[TAG::TABLE::TAGS], data1))
             {
                 emit this->tagRemoved(tag);
                 return true;
             }
         }
-    }    
+    }
     
     return false;
 }
@@ -257,9 +311,9 @@ bool Tagging::removeTag(const QString& tag)
 static bool doNameFilter(const QString &name, const QStringList &filters)
 {
     const auto filtersAccumulate = std::accumulate(filters.constBegin(), filters.constEnd(), QVector<QRegExp> {}, [](QVector<QRegExp> &res, const QString &filter) -> QVector<QRegExp> {
-        res.append(QRegExp(filter, Qt::CaseInsensitive, QRegExp::Wildcard));
-        return res;
-    });
+            res.append(QRegExp(filter, Qt::CaseInsensitive, QRegExp::Wildcard));
+            return res;
+});
     
     for (const auto &filter : filtersAccumulate) {
         if (filter.exactMatch(name)) {
@@ -268,8 +322,8 @@ static bool doNameFilter(const QString &name, const QStringList &filters)
     }
     return false;
 }
-
-QList<QUrl> Tagging::getTagUrls(const QString &tag, const QStringList &filters, const bool &strict, const int &limit, const QString &mime)
+ 
+ QList<QUrl> Tagging::getTagUrls(const QString &tag, const QStringList &filters, const bool &strict, const int &limit, const QString &mime) 
 {
     QList<QUrl> urls;
     
@@ -305,46 +359,48 @@ FMH::MODEL_LIST Tagging::getTags(const int &limit)
         {FMH::MODEL_KEY::IS_DIR, "true"},
         {FMH::MODEL_KEY::LABEL, label},
         {FMH::MODEL_KEY::TYPE, FMStatic::PATHTYPE_LABEL[FMStatic::PATHTYPE_KEY::TAGS_PATH]}};
-    }
-    
-    return data;
+}
+
+return data;
 }
 
 FMH::MODEL_LIST Tagging::getUrlTags(const QUrl &url)
 {
- return FMH::toModelList(getUrlTags(url.toString(), false));   
+    return FMH::toModelList(this->getUrlTags(url.toString(), false));
 }
 
 bool Tagging::addTagToUrl(const QString tag, const QUrl &url)
 {
-   return tagUrl(url.toString(), tag);
+    return this->tagUrl(url.toString(), tag);
 }
 
 bool Tagging::removeTagToUrl(const QString tag, const QUrl &url)
 {
-    return removeUrlTag(url.toString(), tag);
+    return this->removeUrlTag(url.toString(), tag);
 }
 
 bool Tagging::toggleFav(const QUrl &url)
 {
-    if (isFav(url))
-        return unFav(url);
+    if (this->isFav(url))
+        return this->unFav(url);
     
-    return fav(url);
+    return this->fav(url);
 }
 
 bool Tagging::fav(const QUrl &url)
 {
-    return tagUrl(url.toString(), "fav", "#e91e63");
+    return this->tagUrl(url.toString(), "fav", "#e91e63");
 }
 
-bool Tagging::unFav(const QUrl &url)
+bool Tagging::unFav(const QUrl &url) 
 {
-    return removeUrlTag(url.toString(), "fav");
+    return this->removeUrlTag(url.toString(), "fav");
 }
 
 bool Tagging::isFav(const QUrl &url, const bool &strict)
 {
+    Q_UNUSED(strict)
+    
     return urlTagExists(url.toString(), "fav");
 }
 
